#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import os
import shutil
import tempfile
from collections.abc import Callable, Container, Sequence
from functools import cached_property
from typing import TYPE_CHECKING, Any, cast

from airflow.providers.common.compat.sdk import (
    AirflowException,
    AirflowSkipException,
    context_to_airflow_vars,
)
from airflow.providers.standard.hooks.subprocess import SubprocessHook, SubprocessResult, working_directory
from airflow.providers.standard.version_compat import BaseOperator

if TYPE_CHECKING:
    from airflow.providers.common.compat.sdk import Context

    from tests_common.test_utils.version_compat import ArgNotSet


class BashOperator(BaseOperator):
    r"""
    Execute a Bash script, command or set of commands.

    .. seealso::
        For more information on how to use this operator, take a look at the guide:
        :ref:`howto/operator:BashOperator`

    If BaseOperator.do_xcom_push is True, the last line written to stdout
    will also be pushed to an XCom when the bash command completes

    :param bash_command: The command, set of commands or reference to a
        Bash script (must be '.sh' or '.bash') to be executed. (templated)
    :param env: If env is not None, it must be a dict that defines the
        environment variables for the new process; these are used instead
        of inheriting the current process environment, which is the default
        behavior. (templated)
    :param append_env: If False(default) uses the environment variables passed in env params
        and does not inherit the current process environment. If True, inherits the environment variables
        from current passes and then environment variable passed by the user will either update the existing
        inherited environment variables or the new variables gets appended to it
    :param output_encoding: Output encoding of Bash command
    :param skip_on_exit_code: If task exits with this exit code, leave the task
        in ``skipped`` state (default: 99). If set to ``None``, any non-zero
        exit code will be treated as a failure.
    :param cwd: Working directory to execute the command in (templated).
        If None (default), the command is run in a temporary directory.
        To use current DAG folder as the working directory,
        you might set template ``{{ task.dag.folder }}``.
        When bash_command is a '.sh' or '.bash' file, Airflow must have write
        access to the working directory. The script will be rendered (Jinja
        template) into a new temporary file in this directory.
    :param output_processor: Function to further process the output of the bash script
        (default is lambda output: output).

    Airflow will evaluate the exit code of the Bash command. In general, a non-zero exit code will result in
    task failure and zero will result in task success.
    Exit code ``99`` (or another set in ``skip_on_exit_code``)
    will throw an :class:`airflow.exceptions.AirflowSkipException`, which will leave the task in ``skipped``
    state. You can have all non-zero exit codes be treated as a failure by setting ``skip_on_exit_code=None``.

    .. list-table::
       :widths: 25 25
       :header-rows: 1

       * - Exit code
         - Behavior
       * - 0
         - success
       * - `skip_on_exit_code` (default: 99)
         - raise :class:`airflow.exceptions.AirflowSkipException`
       * - otherwise
         - raise :class:`airflow.exceptions.AirflowException`

    .. note::

        Airflow will not recognize a non-zero exit code unless the whole shell exit with a non-zero exit
        code.  This can be an issue if the non-zero exit arises from a sub-command.  The easiest way of
        addressing this is to prefix the command with ``set -e;``

        .. code-block:: python

            bash_command = "set -e; python3 script.py '{{ data_interval_end }}'"

    .. note::

        To simply execute a ``.sh`` or ``.bash`` script (without any Jinja template), add a space after the
        script name ``bash_command`` argument -- for example ``bash_command="my_script.sh "``. This
        is because Airflow tries to load this file and process it as a Jinja template when
        it ends with ``.sh`` or ``.bash``.

        If you have Jinja template in your script, do not put any blank space. And add the script's directory
        in the DAG's ``template_searchpath``. If you specify a ``cwd``, Airflow must have write access to
        this directory. The script will be rendered (Jinja template) into a new temporary file in this directory.

    .. warning::

        Care should be taken with "user" input or when using Jinja templates in the
        ``bash_command``, as this bash operator does not perform any escaping or
        sanitization of the command.

        This applies mostly to using "dag_run" conf, as that can be submitted via
        users in the Web UI. Most of the default template variables are not at
        risk.

    For example, do **not** do this:

    .. code-block:: python

        bash_task = BashOperator(
            task_id="bash_task",
            bash_command='echo "Here is the message: \'{{ dag_run.conf["message"] if dag_run else "" }}\'"',
        )

    Instead, you should pass this via the ``env`` kwarg and use double-quotes
    inside the bash_command, as below:

    .. code-block:: python

        bash_task = BashOperator(
            task_id="bash_task",
            bash_command="echo \"here is the message: '$message'\"",
            env={"message": '{{ dag_run.conf["message"] if dag_run else "" }}'},
        )

    .. versionadded:: 2.10.0
       The `output_processor` parameter.

    """

    template_fields: Sequence[str] = ("bash_command", "env", "cwd")
    template_fields_renderers = {"bash_command": "bash", "env": "json"}
    template_ext: Sequence[str] = (".sh", ".bash")
    ui_color = "#f0ede4"

    def __init__(
        self,
        *,
        bash_command: str | ArgNotSet,
        env: dict[str, str] | None = None,
        append_env: bool = False,
        output_encoding: str = "utf-8",
        skip_on_exit_code: int | Container[int] | None = 99,
        cwd: str | None = None,
        output_processor: Callable[[str], Any] = lambda result: result,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)
        self.bash_command = bash_command
        self.env = env
        self.output_encoding = output_encoding
        self.skip_on_exit_code = (
            skip_on_exit_code
            if isinstance(skip_on_exit_code, Container)
            else [skip_on_exit_code]
            if skip_on_exit_code is not None
            else []
        )
        self.cwd = cwd
        self.append_env = append_env
        self.output_processor = output_processor
        self._is_inline_cmd = None
        if isinstance(bash_command, str):
            self._is_inline_cmd = self._is_inline_command(bash_command=bash_command)

    @cached_property
    def subprocess_hook(self):
        """Returns hook for running the bash command."""
        return SubprocessHook()

    def get_env(self, context) -> dict:
        """Build the set of environment variables to be exposed for the bash command."""
        system_env = os.environ.copy()
        env = self.env
        if env is None:
            env = system_env
        else:
            if self.append_env:
                system_env.update(env)
                env = system_env

        airflow_context_vars = context_to_airflow_vars(context, in_env_var_format=True)
        self.log.debug(
            "Exporting env vars: %s",
            " ".join(f"{k}={v!r}" for k, v in airflow_context_vars.items()),
        )
        env.update(airflow_context_vars)
        return env

    def execute(self, context: Context):
        bash_path: str = shutil.which("bash") or "bash"
        if self.cwd is not None:
            if not os.path.exists(self.cwd):
                raise AirflowException(f"Can not find the cwd: {self.cwd}")
            if not os.path.isdir(self.cwd):
                raise AirflowException(f"The cwd {self.cwd} must be a directory")
        env = self.get_env(context)

        if self._is_inline_cmd:
            result = self._run_inline_command(bash_path=bash_path, env=env)
        else:
            result = self._run_rendered_script_file(bash_path=bash_path, env=env)

        if result.exit_code in self.skip_on_exit_code:
            raise AirflowSkipException(f"Bash command returned exit code {result.exit_code}. Skipping.")
        if result.exit_code != 0:
            raise AirflowException(
                f"Bash command failed. The command returned a non-zero exit code {result.exit_code}."
            )

        return self.output_processor(result.output)

    def _run_inline_command(self, bash_path: str, env: dict) -> SubprocessResult:
        """Pass the bash command as string directly in the subprocess."""
        return self.subprocess_hook.run_command(
            command=[bash_path, "-c", self.bash_command],
            env=env,
            output_encoding=self.output_encoding,
            cwd=self.cwd,
        )

    def _run_rendered_script_file(self, bash_path: str, env: dict) -> SubprocessResult:
        """
        Save the bash command into a file and execute this file.

        This allows for longer commands, and prevents "Argument list too long error".
        """
        with working_directory(cwd=self.cwd) as cwd:
            with tempfile.NamedTemporaryFile(mode="w", dir=cwd, suffix=".sh") as file:
                file.write(cast("str", self.bash_command))
                file.flush()

                bash_script = os.path.basename(file.name)
                return self.subprocess_hook.run_command(
                    command=[bash_path, bash_script],
                    env=env,
                    output_encoding=self.output_encoding,
                    cwd=cwd,
                )

    @classmethod
    def _is_inline_command(cls, bash_command: str) -> bool:
        """Return True if the bash command is an inline string. False if it's a bash script file."""
        return not bash_command.endswith(tuple(cls.template_ext))

    def on_kill(self) -> None:
        self.subprocess_hook.send_sigterm()



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from __future__ import annotations

import pytest

from airflow.providers.common.compat.sdk import AirflowSensorTimeout
from airflow.sensors.base import PokeReturnValue
from airflow.utils.state import State

from tests_common.test_utils.version_compat import AIRFLOW_V_3_0_PLUS

if AIRFLOW_V_3_0_PLUS:
    from airflow.sdk import task
else:
    from airflow.decorators import task  # type: ignore[attr-defined,no-redef]

pytestmark = pytest.mark.db_test


@pytest.mark.skipif(
    not AIRFLOW_V_3_0_PLUS,
    reason="Decorators were part of core not providers, so this test doesnt make sense for < AF3.",
)
class TestSensorDecorator:
    def test_sensor_fails_on_none_python_callable(self):
        not_callable = {}
        with pytest.raises(TypeError):
            task.sensor(not_callable)

    def test_basic_sensor_success(self, run_task):
        sensor_xcom_value = "xcom_value"

        @task.sensor
        def sensor_f():
            return PokeReturnValue(is_done=True, xcom_value=sensor_xcom_value)

        sf = sensor_f()
        run_task(sf.operator)

        assert run_task.state == State.SUCCESS
        assert run_task.xcom.get("return_value") == sensor_xcom_value

    def test_basic_sensor_success_returns_bool(self, run_task):
        @task.sensor
        def sensor_f():
            return True

        sf = sensor_f()
        run_task(sf.operator)

        assert run_task.state == State.SUCCESS

    def test_basic_sensor_failure(self, run_task):
        @task.sensor(timeout=0)
        def sensor_f():
            return PokeReturnValue(is_done=False, xcom_value="xcom_value")

        sf = sensor_f()
        run_task(sf.operator)

        assert run_task.state == State.FAILED
        assert isinstance(run_task.error, AirflowSensorTimeout)

    def test_basic_sensor_failure_returns_bool(self, run_task):
        @task.sensor(timeout=0)
        def sensor_f():
            return PokeReturnValue(is_done=False, xcom_value="xcom_value")

        sf = sensor_f()
        run_task(sf.operator)

        assert run_task.state == State.FAILED
        assert isinstance(run_task.error, AirflowSensorTimeout)

    def test_basic_sensor_soft_fail(self, run_task):
        @task.sensor(timeout=0, soft_fail=True)
        def sensor_f():
            return PokeReturnValue(is_done=False, xcom_value="xcom_value")

        sf = sensor_f()
        run_task(sf.operator)

        assert run_task.state == State.SKIPPED

    def test_basic_sensor_soft_fail_returns_bool(self, run_task):
        @task.sensor(timeout=0, soft_fail=True)
        def sensor_f():
            return False

        sf = sensor_f()
        run_task(sf.operator)

        assert run_task.state == State.SKIPPED

    def test_basic_sensor_get_upstream_output(self, dag_maker, run_task):
        ret_val = 100
        sensor_xcom_value = "xcom_value"

        @task
        def upstream_f() -> int:
            return ret_val

        @task.sensor
        def sensor_f(n: int):
            assert n == ret_val
            return PokeReturnValue(is_done=True, xcom_value=sensor_xcom_value)

        with dag_maker(serialized=True):
            uf = upstream_f()
            sf = sensor_f(uf)

        dr = dag_maker.create_dagrun()
        run_task(uf.operator, dag_id=dr.dag_id, run_id=dr.run_id)
        # Patch the return value for upstream_f
        assert run_task.xcom.get("return_value", task_id="upstream_f") == ret_val
        assert run_task.state == State.SUCCESS

        run_task(sf.operator, dag_id=dr.dag_id, run_id=dr.run_id)
        assert run_task.state == State.SUCCESS
        assert run_task.xcom.get("return_value", task_id="sensor_f") == sensor_xcom_value



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
"""Example DAG demonstrating the usage of the `@task.short_circuit()` TaskFlow decorator."""

from __future__ import annotations

import pendulum

from airflow.providers.common.compat.sdk import TriggerRule
from airflow.providers.standard.operators.empty import EmptyOperator
from airflow.sdk import chain, dag, task


@dag(schedule=None, start_date=pendulum.datetime(2021, 1, 1, tz="UTC"), catchup=False, tags=["example"])
def example_short_circuit_decorator():
    # [START howto_operator_short_circuit]
    @task.short_circuit()
    def check_condition(condition):
        return condition

    ds_true = [EmptyOperator(task_id=f"true_{i}") for i in [1, 2]]
    ds_false = [EmptyOperator(task_id=f"false_{i}") for i in [1, 2]]

    condition_is_true = check_condition.override(task_id="condition_is_true")(condition=True)
    condition_is_false = check_condition.override(task_id="condition_is_false")(condition=False)

    chain(condition_is_true, *ds_true)
    chain(condition_is_false, *ds_false)
    # [END howto_operator_short_circuit]

    # [START howto_operator_short_circuit_trigger_rules]
    [task_1, task_2, task_3, task_4, task_5, task_6] = [
        EmptyOperator(task_id=f"task_{i}") for i in range(1, 7)
    ]

    task_7 = EmptyOperator(task_id="task_7", trigger_rule=TriggerRule.ALL_DONE)

    short_circuit = check_condition.override(task_id="short_circuit", ignore_downstream_trigger_rules=False)(
        condition=False
    )

    chain(task_1, [task_2, short_circuit], [task_3, task_4], [task_5, task_6], task_7)
    # [END howto_operator_short_circuit_trigger_rules]


example_dag = example_short_circuit_decorator()



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from pathlib import Path
from textwrap import dedent
from unittest import mock

import pytest

from airflow.providers.standard.utils.python_virtualenv import _generate_pip_conf, _use_uv, prepare_virtualenv

from tests_common.test_utils.config import conf_vars
from tests_common.test_utils.version_compat import remove_task_decorator


class TestPrepareVirtualenv:
    @mock.patch("shutil.which")
    def test_use_uv(self, mock_shutil_which):
        with conf_vars({("standard", "venv_install_method"): "auto"}):
            mock_shutil_which.side_effect = [True]
            assert _use_uv() is True

            mock_shutil_which.side_effect = [False]
            assert _use_uv() is False

        with conf_vars({("standard", "venv_install_method"): "uv"}):
            assert _use_uv() is True

        with conf_vars({("standard", "venv_install_method"): "pip"}):
            assert _use_uv() is False

    @pytest.mark.parametrize(
        ("index_urls", "expected_pip_conf_content", "unexpected_pip_conf_content"),
        [
            [[], ["[global]", "no-index ="], ["index-url", "extra", "http", "pypi"]],
            [["http://mysite"], ["[global]", "index-url", "http://mysite"], ["no-index", "extra", "pypi"]],
            [
                ["http://mysite", "https://othersite"],
                ["[global]", "index-url", "http://mysite", "extra", "https://othersite"],
                ["no-index", "pypi"],
            ],
            [
                ["http://mysite", "https://othersite", "http://site"],
                ["[global]", "index-url", "http://mysite", "extra", "https://othersite http://site"],
                ["no-index", "pypi"],
            ],
        ],
    )
    def test_generate_pip_conf(
        self,
        index_urls: list[str],
        expected_pip_conf_content: list[str],
        unexpected_pip_conf_content: list[str],
        tmp_path: Path,
    ):
        tmp_file = tmp_path / "pip.conf"
        _generate_pip_conf(tmp_file, index_urls)
        generated_conf = tmp_file.read_text()
        for term in expected_pip_conf_content:
            assert term in generated_conf
        for term in unexpected_pip_conf_content:
            assert term not in generated_conf

    @mock.patch("airflow.providers.standard.utils.python_virtualenv._execute_in_subprocess")
    @conf_vars({("standard", "venv_install_method"): "pip"})
    def test_should_create_virtualenv_pip(self, mock_execute_in_subprocess):
        python_bin = prepare_virtualenv(
            venv_directory="/VENV", python_bin="pythonVER", system_site_packages=False, requirements=[]
        )
        assert python_bin == "/VENV/bin/python"
        mock_execute_in_subprocess.assert_called_once_with(["pythonVER", "-m", "venv", "/VENV"])

    @mock.patch("airflow.providers.standard.utils.python_virtualenv._execute_in_subprocess")
    @conf_vars({("standard", "venv_install_method"): "uv"})
    def test_should_create_virtualenv_uv(self, mock_execute_in_subprocess):
        python_bin = prepare_virtualenv(
            venv_directory="/VENV", python_bin="pythonVER", system_site_packages=False, requirements=[]
        )
        assert python_bin == "/VENV/bin/python"
        mock_execute_in_subprocess.assert_called_once_with(
            ["uv", "venv", "--allow-existing", "--seed", "--python", "pythonVER", "/VENV"],
            env=mock.ANY,
        )

    @mock.patch("airflow.providers.standard.utils.python_virtualenv._execute_in_subprocess")
    @conf_vars({("standard", "venv_install_method"): "pip"})
    def test_should_create_virtualenv_with_system_packages_pip(self, mock_execute_in_subprocess):
        python_bin = prepare_virtualenv(
            venv_directory="/VENV", python_bin="pythonVER", system_site_packages=True, requirements=[]
        )
        assert python_bin == "/VENV/bin/python"
        mock_execute_in_subprocess.assert_called_once_with(
            ["pythonVER", "-m", "venv", "/VENV", "--system-site-packages"]
        )

    @mock.patch("airflow.providers.standard.utils.python_virtualenv._execute_in_subprocess")
    @conf_vars({("standard", "venv_install_method"): "uv"})
    def test_should_create_virtualenv_with_system_packages_uv(self, mock_execute_in_subprocess):
        python_bin = prepare_virtualenv(
            venv_directory="/VENV", python_bin="pythonVER", system_site_packages=True, requirements=[]
        )
        assert python_bin == "/VENV/bin/python"
        mock_execute_in_subprocess.assert_called_once_with(
            [
                "uv",
                "venv",
                "--allow-existing",
                "--seed",
                "--python",
                "pythonVER",
                "--system-site-packages",
                "/VENV",
            ],
            env=mock.ANY,
        )

    @mock.patch("airflow.providers.standard.utils.python_virtualenv._execute_in_subprocess")
    @conf_vars({("standard", "venv_install_method"): "pip"})
    def test_pip_install_options_pip(self, mock_execute_in_subprocess):
        pip_install_options = ["--no-deps"]
        python_bin = prepare_virtualenv(
            venv_directory="/VENV",
            python_bin="pythonVER",
            system_site_packages=True,
            requirements=["apache-beam[gcp]"],
            pip_install_options=pip_install_options,
        )

        assert python_bin == "/VENV/bin/python"
        mock_execute_in_subprocess.assert_called_with(
            ["/VENV/bin/pip", "install", *pip_install_options, "apache-beam[gcp]"],
            env=mock.ANY,
        )

    @mock.patch("airflow.providers.standard.utils.python_virtualenv._execute_in_subprocess")
    @conf_vars({("standard", "venv_install_method"): "uv"})
    def test_pip_install_options_uv(self, mock_execute_in_subprocess):
        pip_install_options = ["--no-deps"]
        python_bin = prepare_virtualenv(
            venv_directory="/VENV",
            python_bin="pythonVER",
            system_site_packages=True,
            requirements=["apache-beam[gcp]"],
            pip_install_options=pip_install_options,
        )

        assert python_bin == "/VENV/bin/python"
        mock_execute_in_subprocess.assert_called_with(
            [
                "uv",
                "pip",
                "install",
                "--python",
                "/VENV/bin/python",
                *pip_install_options,
                "apache-beam[gcp]",
            ],
            env=mock.ANY,
        )

    @mock.patch("airflow.providers.standard.utils.python_virtualenv._execute_in_subprocess")
    @conf_vars({("standard", "venv_install_method"): "pip"})
    def test_should_create_virtualenv_with_extra_packages_pip(self, mock_execute_in_subprocess):
        python_bin = prepare_virtualenv(
            venv_directory="/VENV",
            python_bin="pythonVER",
            system_site_packages=False,
            requirements=["apache-beam[gcp]"],
        )
        assert python_bin == "/VENV/bin/python"

        mock_execute_in_subprocess.assert_any_call(["pythonVER", "-m", "venv", "/VENV"])

        mock_execute_in_subprocess.assert_called_with(
            ["/VENV/bin/pip", "install", "apache-beam[gcp]"], env=mock.ANY
        )

    @mock.patch("airflow.providers.standard.utils.python_virtualenv._execute_in_subprocess")
    @conf_vars({("standard", "venv_install_method"): "uv"})
    def test_should_create_virtualenv_with_extra_packages_uv(self, mock_execute_in_subprocess):
        python_bin = prepare_virtualenv(
            venv_directory="/VENV",
            python_bin="pythonVER",
            system_site_packages=False,
            requirements=["apache-beam[gcp]"],
        )
        assert python_bin == "/VENV/bin/python"

        mock_execute_in_subprocess.assert_called_with(
            ["uv", "pip", "install", "--python", "/VENV/bin/python", "apache-beam[gcp]"],
            env=mock.ANY,
        )

    @mock.patch("airflow.providers.standard.utils.python_virtualenv._generate_pip_conf")
    @mock.patch("airflow.providers.standard.utils.python_virtualenv._execute_in_subprocess")
    @conf_vars({("standard", "venv_install_method"): "uv"})
    def test_venv_creation_with_index_urls(
        self, mock_execute_in_subprocess, mock_generate_pip_conf, tmp_path: Path
    ):
        """
        Test that uv venv creation passes UV_DEFAULT_INDEX env var.

        When package-index connections are available, UV_DEFAULT_INDEX and UV_INDEX
        environment variables are passed to uv venv command to ensure packages can be
        downloaded from the specified index during venv creation (e.g., when installing
        seed packages like pip, setuptools, wheel).
        """
        venv_dir = str(tmp_path / "venv")
        python_bin = prepare_virtualenv(
            venv_directory=venv_dir,
            python_bin="pythonVER",
            system_site_packages=False,
            requirements=["somepackage"],
            index_urls=["https://private.package.index"],
        )
        assert python_bin == f"{venv_dir}/bin/python"

        # First call: venv creation should have UV_DEFAULT_INDEX in env
        venv_call_args = mock_execute_in_subprocess.call_args_list[0]
        venv_cmd = venv_call_args[0][0]
        venv_kwargs = venv_call_args[1]

        assert venv_cmd == ["uv", "venv", "--allow-existing", "--seed", "--python", "pythonVER", venv_dir]
        assert "env" in venv_kwargs
        assert venv_kwargs["env"]["UV_DEFAULT_INDEX"] == "https://private.package.index"

        # Second call: pip install should also have UV_DEFAULT_INDEX
        pip_call_args = mock_execute_in_subprocess.call_args_list[1]
        pip_cmd = pip_call_args[0][0]
        pip_kwargs = pip_call_args[1]

        assert pip_cmd == ["uv", "pip", "install", "--python", f"{venv_dir}/bin/python", "somepackage"]
        assert "env" in pip_kwargs
        assert pip_kwargs["env"]["UV_DEFAULT_INDEX"] == "https://private.package.index"

    @pytest.mark.parametrize(
        ("decorators", "expected_decorators"),
        [
            (["@task.virtualenv"], []),
            (["@task.virtualenv()"], []),
            (['@task.virtualenv(serializer="dill")'], []),
            (["@foo", "@task.virtualenv", "@bar"], ["@foo", "@bar"]),
            (["@foo", "@task.virtualenv()", "@bar"], ["@foo", "@bar"]),
        ],
        ids=["without_parens", "parens", "with_args", "nested_without_parens", "nested_with_parens"],
    )
    def test_remove_task_decorator(self, decorators: list[str], expected_decorators: list[str]):
        concated_decorators = "\n".join(decorators)
        expected_decorator = "\n".join(expected_decorators)
        SCRIPT = dedent(
            """
        def f():
            # @task.virtualenv
            import funcsigs
        """
        )
        py_source = concated_decorators + SCRIPT
        expected_source = expected_decorator + SCRIPT if expected_decorator else SCRIPT.lstrip()

        res = remove_task_decorator(python_source=py_source, task_decorator_name="@task.virtualenv")
        assert res == expected_source



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
"""Test for Package Index Hook."""

from __future__ import annotations

import pytest

from airflow.models.connection import Connection
from airflow.providers.standard.hooks.package_index import PackageIndexHook


class MockConnection(Connection):
    """Mock for the Connection class."""

    def __init__(self, host: str | None, login: str | None, password: str | None):
        super().__init__()
        self.host = host
        self.login = login
        self.password = password


PI_MOCK_TESTDATA = {
    "missing-url": {},
    "anonymous-https": {
        "host": "https://site/path",
        "expected_result": "https://site/path",
    },
    "no_password-http": {
        "host": "http://site/path",
        "login": "any_user",
        "expected_result": "http://any_user@site/path",
    },
    "with_password-http": {
        "host": "http://site/path",
        "login": "any_user",
        "password": "secret@_%1234!",
        "expected_result": "http://any_user:secret%40_%251234%21@site/path",
    },
    "with_password-https": {
        "host": "https://old_user:pass@site/path",
        "login": "any_user",
        "password": "secret@_%1234!",
        "expected_result": "https://any_user:secret%40_%251234%21@site/path",
    },
}


@pytest.fixture(
    params=list(PI_MOCK_TESTDATA.values()),
    ids=list(PI_MOCK_TESTDATA.keys()),
)
def mock_get_connection(monkeypatch: pytest.MonkeyPatch, request: pytest.FixtureRequest) -> str | None:
    """Pytest Fixture."""
    testdata: dict[str, str | None] = request.param
    host: str | None = testdata.get("host", None)
    login: str | None = testdata.get("login", None)
    password: str | None = testdata.get("password", None)
    expected_result: str | None = testdata.get("expected_result", None)
    monkeypatch.setattr(
        "airflow.providers.standard.hooks.package_index.PackageIndexHook.get_connection",
        lambda *_: MockConnection(host, login, password),
    )
    return expected_result


def test_get_connection_url(mock_get_connection: str | None):
    """Test if connection url is assembled correctly from credentials and index_url."""
    expected_result = mock_get_connection
    hook_instance = PackageIndexHook()
    if expected_result:
        connection_url = hook_instance.get_connection_url()
        assert connection_url == expected_result
    else:
        with pytest.raises(ValueError, match="Please provide an index URL."):
            hook_instance.get_connection_url()


@pytest.mark.parametrize("success", [0, 1])
def test_test_connection(monkeypatch: pytest.MonkeyPatch, mock_get_connection: str | None, success: int):
    """Test if connection test responds correctly to return code."""

    def mock_run(*_, **__):
        class MockProc:
            """Mock class."""

            returncode = success
            stderr = "some error text"

        return MockProc()

    monkeypatch.setattr("airflow.providers.standard.hooks.package_index.subprocess.run", mock_run)

    hook_instance = PackageIndexHook()
    if mock_get_connection:
        result = hook_instance.test_connection()
        assert result[0] == (success == 0)
    else:
        with pytest.raises(ValueError, match="Please provide an index URL"):
            hook_instance.test_connection()


def test_get_ui_field_behaviour():
    """Tests UI field result structure"""
    ui_field_behavior = PackageIndexHook.get_ui_field_behaviour()
    assert "hidden_fields" in ui_field_behavior
    assert "relabeling" in ui_field_behavior
    assert "placeholders" in ui_field_behavior



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import asyncio

import pytest

from airflow.providers.standard.triggers.file import FileDeleteTrigger, FileTrigger

from tests_common.test_utils.version_compat import AIRFLOW_V_3_0_PLUS


class TestFileTrigger:
    FILE_PATH = "/files/dags/example_async_file.py"

    def test_serialization(self):
        """Asserts that the trigger correctly serializes its arguments and classpath."""
        trigger = FileTrigger(filepath=self.FILE_PATH, poll_interval=5)
        classpath, kwargs = trigger.serialize()
        assert classpath == "airflow.providers.standard.triggers.file.FileTrigger"
        assert kwargs == {
            "filepath": self.FILE_PATH,
            "poke_interval": 5,
            "recursive": False,
        }

    @pytest.mark.asyncio
    async def test_task_file_trigger(self, tmp_path):
        """Asserts that the trigger only goes off on or after file is found"""
        tmp_dir = tmp_path / "test_dir"
        tmp_dir.mkdir()
        p = tmp_dir / "hello.txt"

        trigger = FileTrigger(
            filepath=str(p.resolve()),
            poke_interval=0.2,
        )

        task = asyncio.create_task(trigger.run().__anext__())
        await asyncio.sleep(0.5)

        # It should not have produced a result
        assert task.done() is False

        p.touch()

        await asyncio.sleep(0.5)
        assert task.done() is True

        # Prevents error when task is destroyed while in "pending" state
        asyncio.get_event_loop().stop()


@pytest.mark.skipif(not AIRFLOW_V_3_0_PLUS, reason="Skip on Airflow < 3.0")
class TestFileDeleteTrigger:
    FILE_PATH = "/files/dags/example_async_file.py"

    def test_serialization(self):
        """Asserts that the trigger correctly serializes its arguments and classpath."""
        trigger = FileDeleteTrigger(filepath=self.FILE_PATH, poll_interval=5)
        classpath, kwargs = trigger.serialize()
        assert classpath == "airflow.providers.standard.triggers.file.FileDeleteTrigger"
        assert kwargs == {
            "filepath": self.FILE_PATH,
            "poke_interval": 5,
        }

    @pytest.mark.asyncio
    async def test_file_delete_trigger(self, tmp_path):
        """Asserts that the trigger goes off on or after file is found and that the files gets deleted."""
        tmp_dir = tmp_path / "test_dir"
        tmp_dir.mkdir()
        p = tmp_dir / "hello.txt"

        trigger = FileDeleteTrigger(
            filepath=str(p.resolve()),
            poke_interval=0.2,
        )

        task = asyncio.create_task(trigger.run().__anext__())
        await asyncio.sleep(0.5)

        # It should not have produced a result
        assert task.done() is False

        p.touch()

        await asyncio.sleep(0.5)
        assert p.exists() is False

        # Prevents error when task is destroyed while in "pending" state
        asyncio.get_event_loop().stop()



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import pytest
from pendulum import datetime

try:
    from airflow.sdk import TriggerRule
except ImportError:
    # Compatibility for Airflow < 3.1
    from airflow.utils.trigger_rule import TriggerRule  # type: ignore[no-redef,attr-defined]
from airflow.utils.state import State

from tests_common.test_utils.version_compat import AIRFLOW_V_3_0_1, AIRFLOW_V_3_0_PLUS

if AIRFLOW_V_3_0_PLUS:
    from airflow.providers.common.compat.sdk import DownstreamTasksSkipped
    from airflow.sdk import task
else:
    from airflow.decorators import task  # type: ignore[attr-defined,no-redef]

pytestmark = pytest.mark.db_test


DEFAULT_DATE = datetime(2022, 8, 17)


@pytest.mark.skipif(AIRFLOW_V_3_0_1, reason="Test doesn't run on AF 3.0.1. Companion test below.")
def test_short_circuit_decorator(dag_maker):
    with dag_maker(serialized=True):

        @task
        def empty(): ...

        @task.short_circuit()
        def short_circuit(condition):
            return condition

        short_circuit_false = short_circuit.override(task_id="short_circuit_false")(condition=False)
        task_1 = empty.override(task_id="task_1")()
        short_circuit_false >> task_1

        short_circuit_true = short_circuit.override(task_id="short_circuit_true")(condition=True)
        task_2 = empty.override(task_id="task_2")()
        short_circuit_true >> task_2

        short_circuit_respect_trigger_rules = short_circuit.override(
            task_id="short_circuit_respect_trigger_rules", ignore_downstream_trigger_rules=False
        )(condition=False)
        task_3 = empty.override(task_id="task_3")()
        task_4 = empty.override(task_id="task_4")()
        task_5 = empty.override(task_id="task_5", trigger_rule=TriggerRule.ALL_DONE)()
        short_circuit_respect_trigger_rules >> [task_3, task_4] >> task_5

    dr = dag_maker.create_dagrun()

    for dag_task in dag_maker.dag.tasks:
        dag_maker.run_ti(dag_task.task_id, dr, ignore_ti_state=True)

    task_state_mapping = {
        "short_circuit_false": State.SUCCESS,
        "task_1": State.SKIPPED,
        "short_circuit_true": State.SUCCESS,
        "task_2": State.SUCCESS,
        "short_circuit_respect_trigger_rules": State.SUCCESS,
        "task_3": State.SKIPPED,
        "task_4": State.SKIPPED,
        "task_5": State.SUCCESS,
    }

    tis = dr.get_task_instances()
    for ti in tis:
        assert ti.state == task_state_mapping[ti.task_id]


@pytest.mark.skipif(not AIRFLOW_V_3_0_1, reason="Test only runs on AF3.0.1")
@pytest.mark.parametrize(("condition", "should_be_skipped"), [(True, False), (False, True)])
def test_short_circuit_decorator_af301(dag_maker, session, condition, should_be_skipped):
    with dag_maker(serialized=True, session=session):

        @task.short_circuit()
        def short_circuit(condition):
            return condition

        @task
        def empty(): ...

        short_circuit = short_circuit.override(task_id="short_circuit")(condition=condition)
        task_1 = empty.override(task_id="task_1")()
        short_circuit >> task_1

    dr = dag_maker.create_dagrun()

    ti_sc = dr.get_task_instance("short_circuit")

    # We only need to assert correct exception is raised for the tasks that should be skipped
    # The actual logic of task execution and setting state is already covered by Task Execution tests
    if should_be_skipped:
        with pytest.raises(DownstreamTasksSkipped) as exc_info:
            ti_sc.run()
        assert exc_info.value.tasks == ["task_1"]
    else:
        ti_sc.run()


@pytest.mark.skipif(not AIRFLOW_V_3_0_1, reason="Test only runs on AF3.0.1")
@pytest.mark.parametrize(
    ("ignore_downstream_trigger_rules", "expected"), [(True, State.SKIPPED), (False, State.SUCCESS)]
)
def test_short_circuit_decorator__ignore_downstream_trigger_rules(
    dag_maker, session, ignore_downstream_trigger_rules, expected
):
    with dag_maker(
        dag_id="test_short_circuit_decorator__ignore_downstream_trigger_rules",
        serialized=True,
        session=session,
    ):

        @task.short_circuit()
        def short_circuit():
            return False

        @task
        def empty(): ...

        short_circuit_respect_trigger_rules = short_circuit.override(
            task_id="short_circuit_respect_trigger_rules",
            ignore_downstream_trigger_rules=ignore_downstream_trigger_rules,
        )()
        task_3 = empty.override(task_id="task_3")()
        task_4 = empty.override(task_id="task_4")()
        task_5 = empty.override(task_id="task_5", trigger_rule=TriggerRule.ALL_DONE)()
        short_circuit_respect_trigger_rules >> [task_3, task_4] >> task_5

    dr = dag_maker.create_dagrun(session=session)
    tis = dr.get_task_instances(session=session)

    # We only need to assert correct exception is raised for the tasks that should be skipped
    # The actual logic of task execution and setting state is already covered by Task Execution tests
    skipped_tasks = set()
    successful_tasks = set()

    for ti in tis:
        try:
            ti.run(session=session)

            # If the task completes successfully without exceptions, add it to the successful_tasks set
            successful_tasks.add(ti.task_id)
        except DownstreamTasksSkipped as dts:
            # `DownstreamTasksSkipped` is raised when the task completes successfully
            # and the downstream tasks are skipped
            successful_tasks.add(ti.task_id)

            # Tasks in dts.tasks are going to be skipped in the Task Execution API
            #   so simulate that here by setting the state to SKIPPED
            skipped_tasks |= set(dts.tasks)

    assert "short_circuit_respect_trigger_rules" in successful_tasks
    assert "task_3" in skipped_tasks
    assert "task_4" in skipped_tasks
    if expected == State.SUCCESS:
        assert "task_5" in successful_tasks
    else:
        assert "task_5" in skipped_tasks


def test_short_circuit_with_multiple_outputs(dag_maker):
    @task.short_circuit(multiple_outputs=True)
    def multiple_output():
        return {"x": 1, "y": 2}

    with dag_maker(serialized=True):
        multiple_output()

    dr = dag_maker.create_dagrun()
    dag_maker.run_ti("multiple_output", dr)
    ti = dr.get_task_instances()[0]
    assert ti.xcom_pull() == {"x": 1, "y": 2}


def test_short_circuit_with_multiple_outputs_and_empty_dict(dag_maker):
    @task.short_circuit(multiple_outputs=True)
    def empty_dict():
        return {}

    with dag_maker(serialized=True):
        empty_dict()

    dr = dag_maker.create_dagrun()
    dag_maker.run_ti("empty_dict", dr)
    ti = dr.get_task_instances()[0]
    assert ti.xcom_pull() == {}



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from collections.abc import Sequence
from subprocess import PIPE, STDOUT, Popen
from tempfile import NamedTemporaryFile, TemporaryDirectory, gettempdir
from typing import TYPE_CHECKING

from airflow.providers.common.compat.sdk import AirflowFailException, BaseSensorOperator

if TYPE_CHECKING:
    from airflow.providers.common.compat.sdk import Context


class BashSensor(BaseSensorOperator):
    """
    Executes a bash command/script.

    Return True if and only if the return code is 0.

    :param bash_command: The command, set of commands or reference to a
        bash script (must be '.sh') to be executed.

    :param env: If env is not None, it must be a mapping that defines the
        environment variables for the new process; these are used instead
        of inheriting the current process environment, which is the default
        behavior. (templated)
    :param output_encoding: output encoding of bash command.
    :param retry_exit_code: If task exits with this code, treat the sensor
        as not-yet-complete and retry the check later according to the
        usual retry/timeout settings. Any other non-zero return code will
        be treated as an error, and cause the sensor to fail. If set to
        ``None`` (the default), any non-zero exit code will cause a retry
        and the task will never raise an error except on time-out.

    .. seealso::
        For more information on how to use this sensor,take a look at the guide:
        :ref:`howto/operator:BashSensor`
    """

    template_fields: Sequence[str] = ("bash_command", "env")

    def __init__(
        self, *, bash_command, env=None, output_encoding="utf-8", retry_exit_code: int | None = None, **kwargs
    ):
        super().__init__(**kwargs)
        self.bash_command = bash_command
        self.env = env
        self.output_encoding = output_encoding
        self.retry_exit_code = retry_exit_code

    def poke(self, context: Context):
        """Execute the bash command in a temporary directory."""
        bash_command = self.bash_command
        self.log.info("Tmp dir root location: %s", gettempdir())
        with (
            TemporaryDirectory(prefix="airflowtmp") as tmp_dir,
            NamedTemporaryFile(dir=tmp_dir, prefix=self.task_id) as f,
        ):
            f.write(bytes(bash_command, "utf_8"))
            f.flush()
            fname = f.name
            script_location = tmp_dir + "/" + fname
            self.log.info("Temporary script location: %s", script_location)
            self.log.info("Running command: %s", bash_command)

            with Popen(
                ["bash", fname],
                stdout=PIPE,
                stderr=STDOUT,
                close_fds=True,
                cwd=tmp_dir,
                env=self.env,
                start_new_session=True,
            ) as resp:
                if resp.stdout:
                    self.log.info("Output:")
                    for line in iter(resp.stdout.readline, b""):
                        self.log.info(line.decode(self.output_encoding).strip())
                resp.wait()
                self.log.info("Command exited with return code %s", resp.returncode)

                # zero code means success, the sensor can go green
                if resp.returncode == 0:
                    return True

                # we have a retry exit code, sensor retries if return code matches, otherwise error
                if self.retry_exit_code is not None:
                    if resp.returncode == self.retry_exit_code:
                        self.log.info("Return code matches retry code, will retry later")
                        return False
                    raise AirflowFailException(f"Command exited with return code {resp.returncode}")

                # backwards compatibility: sensor retries no matter the error code
                self.log.info("Non-zero return code and no retry code set, will retry later")
                return False



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from datetime import timedelta
from typing import Any

import pendulum
import pytest
import time_machine

from airflow.exceptions import AirflowProviderDeprecationWarning
from airflow.models.dag import DAG
from airflow.providers.common.compat.sdk import TaskDeferred
from airflow.providers.standard.sensors.time_delta import (
    TimeDeltaSensor,
    TimeDeltaSensorAsync,
    WaitSensor,
)
from airflow.providers.standard.triggers.temporal import DateTimeTrigger
from airflow.utils.types import DagRunType

from tests_common.test_utils import db
from tests_common.test_utils.version_compat import AIRFLOW_V_3_0_PLUS, AIRFLOW_V_3_2_PLUS, timezone

if AIRFLOW_V_3_2_PLUS:
    from airflow.dag_processing.dagbag import DagBag
else:
    from airflow.models.dagbag import DagBag  # type: ignore[attr-defined, no-redef]

pytestmark = pytest.mark.db_test

DEFAULT_DATE = timezone.datetime(2015, 1, 1)
DEV_NULL = "/dev/null"
TEST_DAG_ID = "unit_tests"

if AIRFLOW_V_3_0_PLUS:
    DEFER_PATH = "airflow.sdk.BaseOperator.defer"
else:
    DEFER_PATH = "airflow.models.baseoperator.BaseOperator.defer"


@pytest.fixture(autouse=True)
def clear_db():
    db.clear_db_dags()
    db.clear_db_runs()


class TestTimedeltaSensor:
    def setup_method(self):
        self.dagbag = DagBag(dag_folder=DEV_NULL, include_examples=False)
        self.dag = DAG(TEST_DAG_ID, schedule=timedelta(days=1), start_date=DEFAULT_DATE)

    def test_timedelta_sensor(self, mocker):
        op = TimeDeltaSensor(task_id="timedelta_sensor_check", delta=timedelta(seconds=2), dag=self.dag)
        op.execute({"dag_run": mocker.MagicMock(run_after=DEFAULT_DATE), "data_interval_end": DEFAULT_DATE})


@pytest.mark.parametrize(
    ("run_after", "interval_end"),
    [
        (timezone.utcnow() + timedelta(days=1), timezone.utcnow() + timedelta(days=2)),
        (timezone.utcnow() + timedelta(days=1), None),
    ],
)
def test_timedelta_sensor_run_after_vs_interval(run_after, interval_end, dag_maker, session):
    """Interval end should be used as base time when present else run_after"""
    if not AIRFLOW_V_3_0_PLUS and not interval_end:
        pytest.skip("not applicable")

    context = {}
    if interval_end:
        context["data_interval_end"] = interval_end
    delta = timedelta(seconds=1)
    with dag_maker():
        op = TimeDeltaSensor(task_id="wait_sensor_check", delta=delta, mode="reschedule")

    kwargs = {}
    if AIRFLOW_V_3_0_PLUS:
        from airflow.utils.types import DagRunTriggeredByType

        kwargs.update(triggered_by=DagRunTriggeredByType.TEST, run_after=run_after)
    dr = dag_maker.create_dagrun(
        run_id="abcrhroceuh",
        run_type=DagRunType.MANUAL,
        state=None,
        session=session,
        **kwargs,
    )
    ti = dr.task_instances[0]
    context.update(dag_run=dr, ti=ti)
    expected = interval_end or run_after
    actual = op._derive_base_time(context)
    assert actual == expected


@pytest.mark.parametrize(
    ("run_after", "interval_end"),
    [
        (timezone.utcnow() + timedelta(days=1), timezone.utcnow() + timedelta(days=2)),
        (timezone.utcnow() + timedelta(days=1), None),
    ],
)
def test_timedelta_sensor_deferrable_run_after_vs_interval(run_after, interval_end, dag_maker):
    """Test that TimeDeltaSensor defers correctly when flag is enabled."""
    if not AIRFLOW_V_3_0_PLUS and not interval_end:
        pytest.skip("not applicable")

    context: dict[str, Any] = {}
    if interval_end:
        context["data_interval_end"] = interval_end

    with dag_maker():
        kwargs = {}
        if AIRFLOW_V_3_0_PLUS:
            from airflow.utils.types import DagRunTriggeredByType

            kwargs.update(triggered_by=DagRunTriggeredByType.TEST, run_after=run_after)

        delta = timedelta(minutes=5)
        sensor = TimeDeltaSensor(
            task_id="timedelta_sensor_deferrable",
            delta=delta,
            deferrable=True,  # <-- the feature under test
        )

    dr = dag_maker.create_dagrun(
        run_id="abcrhroceuh",
        run_type=DagRunType.MANUAL,
        state=None,
        **kwargs,
    )
    context.update(dag_run=dr)

    expected_base = interval_end or run_after
    expected_fire_time = expected_base + delta

    with pytest.raises(TaskDeferred) as td:
        sensor.execute(context)

    # The sensor should defer once with a DateTimeTrigger
    trigger = td.value.trigger
    assert isinstance(trigger, DateTimeTrigger)
    assert trigger.moment == expected_fire_time


class TestTimeDeltaSensorAsync:
    def setup_method(self):
        self.dagbag = DagBag(dag_folder=DEV_NULL, include_examples=True)
        self.args = {"owner": "airflow", "start_date": DEFAULT_DATE}
        self.dag = DAG(TEST_DAG_ID, schedule=timedelta(days=1), default_args=self.args)

    @pytest.mark.parametrize(
        "should_defer",
        [False, True],
    )
    def test_timedelta_sensor(self, mocker, should_defer):
        defer_mock = mocker.patch(DEFER_PATH)
        delta = timedelta(hours=1)
        with pytest.warns(AirflowProviderDeprecationWarning):
            op = TimeDeltaSensorAsync(task_id="timedelta_sensor_check", delta=delta, dag=self.dag)
        if should_defer:
            data_interval_end = pendulum.now("UTC").add(hours=1)
        else:
            data_interval_end = pendulum.now("UTC").replace(microsecond=0, second=0, minute=0).add(hours=-1)
        op.execute({"data_interval_end": data_interval_end})
        if should_defer:
            defer_mock.assert_called_once()
        else:
            defer_mock.assert_not_called()

    @pytest.mark.parametrize(
        "should_defer",
        [False, True],
    )
    def test_wait_sensor(self, mocker, should_defer):
        defer_mock = mocker.patch(DEFER_PATH)
        sleep_mock = mocker.patch("airflow.providers.standard.sensors.time_delta.sleep")
        wait_time = timedelta(seconds=30)
        op = WaitSensor(
            task_id="wait_sensor_check", time_to_wait=wait_time, dag=self.dag, deferrable=should_defer
        )
        with time_machine.travel(pendulum.datetime(year=2024, month=8, day=1, tz="UTC"), tick=False):
            op.execute({})
            if should_defer:
                defer_mock.assert_called_once()
            else:
                defer_mock.assert_not_called()
                sleep_mock.assert_called_once_with(30)

    @pytest.mark.parametrize(
        ("run_after", "interval_end"),
        [
            (timezone.utcnow() + timedelta(days=1), timezone.utcnow() + timedelta(days=2)),
            (timezone.utcnow() + timedelta(days=1), None),
        ],
    )
    def test_timedelta_sensor_async_run_after_vs_interval(self, run_after, interval_end, dag_maker):
        """Interval end should be used as base time when present else run_after"""
        if not AIRFLOW_V_3_0_PLUS and not interval_end:
            pytest.skip("not applicable")

        context = {}
        if interval_end:
            context["data_interval_end"] = interval_end
        with dag_maker() as dag:
            ...
        kwargs = {}
        if AIRFLOW_V_3_0_PLUS:
            from airflow.utils.types import DagRunTriggeredByType

            kwargs.update(triggered_by=DagRunTriggeredByType.TEST, run_after=run_after)

        dr = dag_maker.create_dagrun(
            run_id="abcrhroceuh",
            run_type=DagRunType.MANUAL,
            state=None,
            **kwargs,
        )
        context.update(dag_run=dr)
        delta = timedelta(seconds=1)
        with pytest.warns(AirflowProviderDeprecationWarning):
            op = TimeDeltaSensorAsync(task_id="wait_sensor_check", delta=delta, dag=dag)
        base_time = interval_end or run_after
        expected_time = base_time + delta
        with pytest.raises(TaskDeferred) as caught:
            op.execute(context)

        assert caught.value.trigger.moment == expected_time



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from datetime import datetime, time

import pendulum
import pytest
import time_machine

from airflow.models.dag import DAG
from airflow.providers.common.compat.sdk import TaskDeferred
from airflow.providers.standard.sensors.time import TimeSensor
from airflow.providers.standard.triggers.temporal import DateTimeTrigger

from tests_common.test_utils.compat import timezone

DEFAULT_TIMEZONE = pendulum.timezone("Asia/Singapore")  # UTC+08:00
DEFAULT_DATE_WO_TZ = datetime(2015, 1, 1)
DEFAULT_DATE_WITH_TZ = datetime(2015, 1, 1, tzinfo=DEFAULT_TIMEZONE)


class TestTimeSensor:
    @pytest.mark.parametrize(
        ("tzinfo", "start_date", "target_time", "expected"),
        [
            (timezone.utc, DEFAULT_DATE_WO_TZ, time(10, 0), True),
            (timezone.utc, DEFAULT_DATE_WITH_TZ, time(16, 0), True),
            (timezone.utc, DEFAULT_DATE_WITH_TZ, time(23, 0), False),
            (DEFAULT_TIMEZONE, DEFAULT_DATE_WO_TZ, time(23, 0), False),
        ],
    )
    @time_machine.travel(timezone.datetime(2020, 1, 1, 13, 0).replace(tzinfo=timezone.utc))
    def test_timezone(self, tzinfo, start_date, target_time, expected, monkeypatch):
        monkeypatch.setattr("airflow.settings.TIMEZONE", tzinfo)
        dag = DAG("test_timezone", schedule=None, default_args={"start_date": start_date})
        op = TimeSensor(task_id="test", target_time=target_time, dag=dag)
        assert op.poke(None) == expected

    def test_target_time_aware_dag_timezone(self):
        # This behavior should be the same for both deferrable and non-deferrable
        with DAG("test_target_time_aware", schedule=None, start_date=datetime(2020, 1, 1, 13, 0)):
            aware_time = time(0, 1).replace(tzinfo=DEFAULT_TIMEZONE)
            op = TimeSensor(task_id="test", target_time=aware_time)
            assert op.target_datetime.tzinfo == timezone.utc

    @pytest.mark.parametrize(
        ("current_datetime", "server_timezone"),
        [
            ("2025-01-26 22:00:00", "UTC"),
            ("2025-01-27 07:00:00", "Asia/Seoul"),  # UTC+09:00
        ],
    )
    def test_target_date_aware_dag_timezone(self, current_datetime, server_timezone):
        travel_time = pendulum.parse(current_datetime, tz=pendulum.timezone(server_timezone))
        user_timezone = pendulum.timezone("Asia/Seoul")
        expected_target_datetime = pendulum.datetime(2025, 1, 26, 22, 0, 0, tz="UTC")

        with time_machine.travel(travel_time, tick=False):
            with DAG(
                "test_target_date_aware",
                schedule=None,
                start_date=datetime(2025, 1, 27, 7, tzinfo=user_timezone),
            ):
                aware_datetime = pendulum.datetime(2025, 1, 27, 7).replace(tzinfo=user_timezone)
                op = TimeSensor(task_id="test", target_time=aware_datetime.time())

                # In the old logic, if server's timezone is UTC op.target_datetime could be incorrectly.
                # For example, it might be set to `2025-01-25 22:00:00 UTC`, not `2025-01-26 22:00:00 UTC`.
                # This issue stems from using datetime.today() in an environment where the server timezone differs from the user's timezone
                # The local date '2025-01-26' is combined with the target time `07:00:00`,
                # resulting in `2025-01-26 07:00:00` in local time.
                # When this is converted to UTC by convert_to_utc, it becomes `2025-01-25 22:00:00 UTC`.
                assert op.target_datetime == expected_target_datetime

    def test_target_time_naive_dag_timezone(self):
        # Again, this behavior should be the same for both deferrable and non-deferrable
        with DAG(
            dag_id="test_target_time_naive_dag_timezone",
            schedule=None,
            start_date=datetime(2020, 1, 1, 23, 0, tzinfo=DEFAULT_TIMEZONE),
        ):
            op = TimeSensor(task_id="test", target_time=time(9, 0))

            # Since the DEFAULT_TIMEZONE is UTC+8:00, then hour 9 should be converted to hour 1
            assert op.target_datetime.time() == pendulum.time(1, 0)
            assert op.target_datetime.tzinfo == timezone.utc

    @time_machine.travel("2020-07-07 00:00:00", tick=False)
    def test_task_is_deferred(self):
        with DAG(
            dag_id="test_task_is_deferred",
            schedule=None,
            start_date=datetime(2020, 1, 1, 13, 0),
        ):
            op = TimeSensor(task_id="test", target_time=time(10, 0), deferrable=True)

        # This should be converted to UTC in the __init__. Since there is no default timezone, it will become
        # aware, but note changed
        assert op.target_datetime.utcoffset() is not None

        with pytest.raises(TaskDeferred) as exc_info:
            op.execute({})

        assert isinstance(exc_info.value.trigger, DateTimeTrigger)
        assert exc_info.value.trigger.moment == pendulum.datetime(2020, 7, 7, 10)
        assert exc_info.value.kwargs is None
        assert exc_info.value.method_name == "execute_complete"

    def test_execute_complete_accepts_event(self):
        """Ensure execute_complete supports the 'event' kwarg when deferrable=True."""
        with DAG(
            dag_id="test_execute_complete_accepts_event",
            schedule=None,
            start_date=datetime(2020, 1, 1),  # Matches above
        ):
            op = TimeSensor(task_id="test", target_time=time(10, 0), deferrable=True)

        try:
            op.execute_complete(context={}, event={"status": "success"})
        except TypeError as e:
            pytest.fail(f"TypeError raised: {e}")



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import contextlib

import pytest

from airflow.providers.standard.decorators.stub import stub


def fn_ellipsis(): ...


def fn_pass(): ...


def fn_doc():
    """Some string"""


def fn_doc_pass():
    """Some string"""
    pass


def fn_code():
    return None


@pytest.mark.parametrize(
    ("fn", "error"),
    [
        pytest.param(fn_ellipsis, contextlib.nullcontext(), id="ellipsis"),
        pytest.param(fn_pass, contextlib.nullcontext(), id="pass"),
        pytest.param(fn_doc, contextlib.nullcontext(), id="doc"),
        pytest.param(fn_doc_pass, contextlib.nullcontext(), id="doc-and-pass"),
        pytest.param(fn_code, pytest.raises(ValueError, match="must be an empty function"), id="not-empty"),
    ],
)
def test_stub_signature(fn, error):
    with error:
        stub(fn)()



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from __future__ import annotations

import pendulum

from airflow.providers.common.compat.sdk import TriggerRule
from airflow.providers.standard.operators.empty import EmptyOperator
from airflow.providers.standard.utils.weekday import WeekDay
from airflow.sdk import chain, dag, task
from airflow.sdk.exceptions import AirflowSkipException


@dag(schedule=None, start_date=pendulum.datetime(2023, 1, 1, tz="UTC"), catchup=False)
def example_bash_decorator():
    @task.bash
    def run_me(sleep_seconds: int, task_instance_key_str: str) -> str:
        return f"echo {task_instance_key_str} && sleep {sleep_seconds}"

    run_me_loop = [run_me.override(task_id=f"runme_{i}")(sleep_seconds=i) for i in range(3)]

    # [START howto_decorator_bash]
    @task.bash
    def run_after_loop() -> str:
        return "echo https://airflow.apache.org/"

    run_this = run_after_loop()
    # [END howto_decorator_bash]

    # [START howto_decorator_bash_template]
    @task.bash
    def also_run_this() -> str:
        return 'echo "ti_key={{ task_instance_key_str }}"'

    also_this = also_run_this()
    # [END howto_decorator_bash_template]

    # [START howto_decorator_bash_context_vars]
    @task.bash
    def also_run_this_again(task_instance_key_str) -> str:
        return f'echo "ti_key={task_instance_key_str}"'

    also_this_again = also_run_this_again()
    # [END howto_decorator_bash_context_vars]

    # [START howto_decorator_bash_skip]
    @task.bash
    def this_will_skip() -> str:
        return 'echo "hello world"; exit 99;'

    this_skips = this_will_skip()
    # [END howto_decorator_bash_skip]

    run_this_last = EmptyOperator(task_id="run_this_last", trigger_rule=TriggerRule.ALL_DONE)

    # [START howto_decorator_bash_conditional]
    @task.bash
    def sleep_in(day: str) -> str:
        if day in (WeekDay.SATURDAY, WeekDay.SUNDAY):
            return f"sleep {60 * 60}"
        raise AirflowSkipException("No sleeping in today!")

    sleep_in(day="{{ dag_run.logical_date.strftime('%A').lower() }}")
    # [END howto_decorator_bash_conditional]

    # [START howto_decorator_bash_parametrize]
    @task.bash(env={"BASE_DIR": "{{ dag_run.logical_date.strftime('%Y/%m/%d') }}"}, append_env=True)
    def make_dynamic_dirs(new_dirs: str) -> str:
        return f"mkdir -p $AIRFLOW_HOME/$BASE_DIR/{new_dirs}"

    make_dynamic_dirs(new_dirs="foo/bar/baz")
    # [END howto_decorator_bash_parametrize]

    # [START howto_decorator_bash_build_cmd]
    @task.bash
    def get_file_stats() -> str:
        from pathlib import Path
        from shlex import join

        # Get stats of the current DAG file itself
        current_file = str(Path(__file__))
        cmd = join(["stat", current_file])

        return cmd

    get_file_stats()
    # [END howto_decorator_bash_build_cmd]

    chain(run_me_loop, run_this)
    chain([also_this, also_this_again, this_skips, run_this], run_this_last)


example_bash_decorator()



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import signal
from pathlib import Path
from subprocess import PIPE, STDOUT
from tempfile import TemporaryDirectory
from unittest import mock
from unittest.mock import MagicMock

import pytest

from airflow.providers.standard.hooks.subprocess import SubprocessHook

OS_ENV_KEY = "SUBPROCESS_ENV_TEST"
OS_ENV_VAL = "this-is-from-os-environ"


class TestSubprocessHook:
    @pytest.mark.parametrize(
        ("env", "expected"),
        [
            ({"ABC": "123", "AAA": "456"}, {"ABC": "123", "AAA": "456", OS_ENV_KEY: ""}),
            ({}, {OS_ENV_KEY: ""}),
            (None, {OS_ENV_KEY: OS_ENV_VAL}),
        ],
        ids=["with env", "empty env", "no env"],
    )
    def test_env(self, env, expected):
        """
        Test that env variables are exported correctly to the command environment.
        When ``env`` is ``None``, ``os.environ`` should be passed to ``Popen``.
        Otherwise, the variables in ``env`` should be available, and ``os.environ`` should not.
        """
        hook = SubprocessHook()

        def build_cmd(keys, filename):
            """
            Produce bash command to echo env vars into filename.
            Will always echo the special test var named ``OS_ENV_KEY`` into the file to test whether
            ``os.environ`` is passed or not.
            """
            return "\n".join(f"echo {k}=${k}>> {filename}" for k in [*keys, OS_ENV_KEY])

        with TemporaryDirectory() as tmp_dir, mock.patch.dict("os.environ", {OS_ENV_KEY: OS_ENV_VAL}):
            tmp_file = Path(tmp_dir, "test.txt")
            command = build_cmd(env and env.keys() or [], tmp_file.as_posix())
            hook.run_command(command=["bash", "-c", command], env=env)
            actual = dict([x.split("=") for x in tmp_file.read_text().splitlines()])
            assert actual == expected

    @pytest.mark.parametrize(
        ("val", "expected"),
        [
            ("test-val", "test-val"),
            ("test-val\ntest-val\n", ""),
            ("test-val\ntest-val", "test-val"),
            ("", ""),
        ],
    )
    def test_return_value(self, val, expected):
        hook = SubprocessHook()
        result = hook.run_command(command=["bash", "-c", f'echo "{val}"'])
        assert result.output == expected

    @mock.patch.dict("os.environ", clear=True)
    @mock.patch(
        "airflow.providers.standard.hooks.subprocess.TemporaryDirectory",
        return_value=MagicMock(__enter__=MagicMock(return_value="/tmp/airflowtmpcatcat")),
    )
    @mock.patch(
        "airflow.providers.standard.hooks.subprocess.Popen",
        return_value=MagicMock(stdout=MagicMock(readline=MagicMock(side_effect=StopIteration), returncode=0)),
    )
    def test_should_exec_subprocess(self, mock_popen, mock_temporary_directory):
        hook = SubprocessHook()
        hook.run_command(command=["bash", "-c", 'echo "stdout"'])

        mock_temporary_directory.assert_called_once()
        mock_popen.assert_called_once_with(
            ["bash", "-c", 'echo "stdout"'],
            cwd="/tmp/airflowtmpcatcat",
            env={},
            restore_signals=True,
            start_new_session=True,
            stderr=STDOUT,
            stdout=PIPE,
        )

    def test_task_decode(self):
        hook = SubprocessHook()
        command = ["bash", "-c", 'printf "This will cause a coding error \\xb1\\xa6\\x01\n"']
        result = hook.run_command(command=command)
        assert result.exit_code == 0

    @mock.patch("os.getpgid", return_value=123)
    @mock.patch("os.killpg")
    def test_send_sigterm(self, mock_killpg, mock_getpgid):
        hook = SubprocessHook()
        hook.sub_process = MagicMock()
        hook.send_sigterm()
        mock_getpgid.assert_called_once()
        mock_killpg.assert_called_with(123, signal.SIGTERM)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import datetime
import os
import typing
import warnings
from collections.abc import Callable, Collection, Iterable, Sequence
from typing import TYPE_CHECKING, ClassVar

from airflow.models.dag import DagModel
from airflow.providers.common.compat.sdk import (
    AirflowSkipException,
    BaseOperatorLink,
    BaseSensorOperator,
    conf,
)
from airflow.providers.standard.exceptions import (
    DuplicateStateError,
    ExternalDagDeletedError,
    ExternalDagFailedError,
    ExternalDagNotFoundError,
    ExternalTaskFailedError,
    ExternalTaskGroupFailedError,
    ExternalTaskGroupNotFoundError,
    ExternalTaskNotFoundError,
)
from airflow.providers.standard.operators.empty import EmptyOperator
from airflow.providers.standard.triggers.external_task import WorkflowTrigger
from airflow.providers.standard.utils.sensor_helper import _get_count, _get_external_task_group_task_ids
from airflow.providers.standard.version_compat import (
    AIRFLOW_V_3_0_PLUS,
    AIRFLOW_V_3_2_PLUS,
    BaseOperator,
)
from airflow.utils.file import correct_maybe_zipped
from airflow.utils.state import State, TaskInstanceState

if not AIRFLOW_V_3_0_PLUS:
    from airflow.utils.session import NEW_SESSION, provide_session

if AIRFLOW_V_3_2_PLUS:
    from airflow.dag_processing.dagbag import DagBag
else:
    from airflow.models.dagbag import DagBag  # type: ignore[attr-defined, no-redef]

if TYPE_CHECKING:
    from sqlalchemy.orm import Session

    from airflow.providers.common.compat.sdk import Context, TaskInstanceKey


class ExternalDagLink(BaseOperatorLink):
    """
    Operator link for ExternalTaskSensor and ExternalTaskMarker.

    It allows users to access DAG waited with ExternalTaskSensor or cleared by ExternalTaskMarker.
    """

    name = "External DAG"

    def get_link(self, operator: BaseOperator, *, ti_key: TaskInstanceKey) -> str:
        if TYPE_CHECKING:
            assert isinstance(operator, (ExternalTaskMarker, ExternalTaskSensor))

        external_dag_id = operator.external_dag_id

        if not AIRFLOW_V_3_0_PLUS:
            from airflow.models.renderedtifields import RenderedTaskInstanceFields
            from airflow.models.taskinstancekey import TaskInstanceKey as CoreTaskInstanceKey

            core_ti_key = CoreTaskInstanceKey(
                dag_id=ti_key.dag_id,
                task_id=ti_key.task_id,
                run_id=ti_key.run_id,
                try_number=ti_key.try_number,
                map_index=ti_key.map_index,
            )

            if template_fields := RenderedTaskInstanceFields.get_templated_fields(core_ti_key):
                external_dag_id: str = template_fields.get("external_dag_id", operator.external_dag_id)  # type: ignore[no-redef]

        if AIRFLOW_V_3_0_PLUS:
            from airflow.utils.helpers import build_airflow_dagrun_url

            return build_airflow_dagrun_url(dag_id=external_dag_id, run_id=ti_key.run_id)
        from airflow.utils.helpers import build_airflow_url_with_query  # type:ignore[attr-defined]

        query = {"dag_id": external_dag_id, "run_id": ti_key.run_id}
        return build_airflow_url_with_query(query)


class ExternalTaskSensor(BaseSensorOperator):
    """
    Waits for a different DAG, task group, or task to complete for a specific logical date.

    If both `external_task_group_id` and `external_task_id` are ``None`` (default), the sensor
    waits for the DAG.
    Values for `external_task_group_id` and `external_task_id` can't be set at the same time.

    By default, the ExternalTaskSensor will wait for the external task to
    succeed, at which point it will also succeed. However, by default it will
    *not* fail if the external task fails, but will continue to check the status
    until the sensor times out (thus giving you time to retry the external task
    without also having to clear the sensor).

    By default, the ExternalTaskSensor will not skip if the external task skips.
    To change this, simply set ``skipped_states=[TaskInstanceState.SKIPPED]``.
    Note that if you are monitoring multiple tasks, and one enters error state
    and the other enters a skipped state, then the external task will react to
    whichever one it sees first. If both happen together, then the failed state
    takes priority.

    It is possible to alter the default behavior by setting states which
    cause the sensor to fail, e.g. by setting ``allowed_states=[DagRunState.FAILED]``
    and ``failed_states=[DagRunState.SUCCESS]`` you will flip the behaviour to
    get a sensor which goes green when the external task *fails* and immediately
    goes red if the external task *succeeds*!

    Note that ``soft_fail`` is respected when examining the failed_states. Thus
    if the external task enters a failed state and ``soft_fail == True`` the
    sensor will _skip_ rather than fail. As a result, setting ``soft_fail=True``
    and ``failed_states=[DagRunState.SKIPPED]`` will result in the sensor
    skipping if the external task skips. However, this is a contrived
    example---consider using ``skipped_states`` if you would like this
    behaviour. Using ``skipped_states`` allows the sensor to skip if the target
    fails, but still enter failed state on timeout. Using ``soft_fail == True``
    as above will cause the sensor to skip if the target fails, but also if it
    times out.

    :param external_dag_id: The dag_id that contains the task you want to
        wait for. (templated)
    :param external_task_id: The task_id that contains the task you want to
        wait for. (templated)
    :param external_task_ids: The list of task_ids that you want to wait for. (templated)
        If ``None`` (default value) the sensor waits for the DAG. Either
        external_task_id or external_task_ids can be passed to
        ExternalTaskSensor, but not both.
    :param external_task_group_id: The task_group_id that contains the task you want to
        wait for. (templated)
    :param allowed_states: Iterable of allowed states, default is ``['success']``
    :param skipped_states: Iterable of states to make this task mark as skipped, default is ``None``
    :param failed_states: Iterable of failed or dis-allowed states, default is ``None``
    :param execution_delta: time difference with the previous execution to
        look at, the default is the same logical date as the current task or DAG.
        For yesterday, use [positive!] datetime.timedelta(days=1). Either
        execution_delta or execution_date_fn can be passed to
        ExternalTaskSensor, but not both.
    :param execution_date_fn: function that receives the current execution's logical date as the first
        positional argument and optionally any number of keyword arguments available in the
        context dictionary, and returns the desired logical dates to query.
        Either execution_delta or execution_date_fn can be passed to ExternalTaskSensor,
        but not both.
    :param check_existence: Set to `True` to check if the external task exists (when
        external_task_id is not None) or check if the DAG to wait for exists (when
        external_task_id is None), and immediately cease waiting if the external task
        or DAG does not exist (default value: False).
    :param poll_interval: polling period in seconds to check for the status
    :param deferrable: Run sensor in deferrable mode
    """

    template_fields = ["external_dag_id", "external_task_id", "external_task_ids", "external_task_group_id"]
    ui_color = "#4db7db"
    operator_extra_links = [ExternalDagLink()]

    def __init__(
        self,
        *,
        external_dag_id: str,
        external_task_id: str | None = None,
        external_task_ids: Collection[str] | None = None,
        external_task_group_id: str | None = None,
        allowed_states: Iterable[str] | None = None,
        skipped_states: Iterable[str] | None = None,
        failed_states: Iterable[str] | None = None,
        execution_delta: datetime.timedelta | None = None,
        execution_date_fn: Callable | None = None,
        check_existence: bool = False,
        poll_interval: float = 2.0,
        deferrable: bool = conf.getboolean("operators", "default_deferrable", fallback=False),
        **kwargs,
    ):
        super().__init__(**kwargs)

        self.allowed_states = list(allowed_states) if allowed_states else [TaskInstanceState.SUCCESS.value]
        self.skipped_states = list(skipped_states) if skipped_states else []
        self.failed_states = list(failed_states) if failed_states else []

        total_states = set(self.allowed_states + self.skipped_states + self.failed_states)

        if len(total_states) != len(self.allowed_states) + len(self.skipped_states) + len(self.failed_states):
            raise DuplicateStateError(
                "Duplicate values provided across allowed_states, skipped_states and failed_states."
            )

        # convert [] to None
        if not external_task_ids:
            external_task_ids = None

        # can't set both single task id and a list of task ids
        if external_task_id is not None and external_task_ids is not None:
            raise ValueError(
                "Only one of `external_task_id` or `external_task_ids` may "
                "be provided to ExternalTaskSensor; "
                "use external_task_id or external_task_ids or external_task_group_id."
            )

        # since both not set, convert the single id to a 1-elt list - from here on, we only consider the list
        if external_task_id is not None:
            external_task_ids = [external_task_id]

        if external_task_group_id is not None and external_task_ids is not None:
            raise ValueError(
                "Only one of `external_task_group_id` or `external_task_ids` may "
                "be provided to ExternalTaskSensor; "
                "use external_task_id or external_task_ids or external_task_group_id."
            )

        # check the requested states are all valid states for the target type, be it dag or task
        if external_task_ids or external_task_group_id:
            if not total_states <= set(State.task_states):
                raise ValueError(
                    "Valid values for `allowed_states`, `skipped_states` and `failed_states` "
                    "when `external_task_id` or `external_task_ids` or `external_task_group_id` "
                    f"is not `None`: {State.task_states}"
                )

        elif not total_states <= set(State.dag_states):
            raise ValueError(
                "Valid values for `allowed_states`, `skipped_states` and `failed_states` "
                f"when `external_task_id` and `external_task_group_id` is `None`: {State.dag_states}"
            )

        if execution_delta is not None and execution_date_fn is not None:
            raise ValueError(
                "Only one of `execution_delta` or `execution_date_fn` may "
                "be provided to ExternalTaskSensor; not both."
            )

        self.execution_delta = execution_delta
        self.execution_date_fn = execution_date_fn
        self.external_dag_id = external_dag_id
        self.external_task_id = external_task_id
        self.external_task_ids = external_task_ids
        self.external_task_group_id = external_task_group_id
        self.check_existence = check_existence
        self._has_checked_existence = False
        self.deferrable = deferrable
        self.poll_interval = poll_interval
        self.external_dates_filter: str | None = None

    def _get_dttm_filter(self, context: Context) -> Sequence[datetime.datetime]:
        logical_date = self._get_logical_date(context)

        if self.execution_delta:
            return [logical_date - self.execution_delta]
        if self.execution_date_fn:
            result = self._handle_execution_date_fn(context=context)
            return result if isinstance(result, list) else [result]
        return [logical_date]

    @staticmethod
    def _serialize_dttm_filter(dttm_filter: Sequence[datetime.datetime]) -> str:
        return ",".join(dt.isoformat() for dt in dttm_filter)

    def poke(self, context: Context) -> bool:
        # delay check to poke rather than __init__ in case it was supplied as XComArgs
        if self.external_task_ids and len(self.external_task_ids) > len(set(self.external_task_ids)):
            raise ValueError("Duplicate task_ids passed in external_task_ids parameter")

        dttm_filter = self._get_dttm_filter(context)
        serialized_dttm_filter = self._serialize_dttm_filter(dttm_filter)
        # Save as attribute - to be used by listeners
        self.external_dates_filter = serialized_dttm_filter

        if self.external_task_ids:
            self.log.info(
                "Poking for tasks %s in dag %s on %s ... ",
                self.external_task_ids,
                self.external_dag_id,
                serialized_dttm_filter,
            )

        if self.external_task_group_id:
            self.log.info(
                "Poking for task_group '%s' in dag '%s' on %s ... ",
                self.external_task_group_id,
                self.external_dag_id,
                serialized_dttm_filter,
            )

        if self.external_dag_id and not self.external_task_group_id and not self.external_task_ids:
            self.log.info(
                "Poking for DAG '%s' on %s ... ",
                self.external_dag_id,
                serialized_dttm_filter,
            )

        if AIRFLOW_V_3_0_PLUS:
            return self._poke_af3(context, dttm_filter)
        return self._poke_af2(dttm_filter)

    def _poke_af3(self, context: Context, dttm_filter: Sequence[datetime.datetime]) -> bool:
        from airflow.providers.standard.utils.sensor_helper import _get_count_by_matched_states

        self._has_checked_existence = True
        ti = context["ti"]

        def _get_count(states: list[str]) -> int:
            if self.external_task_ids:
                return ti.get_ti_count(
                    dag_id=self.external_dag_id,
                    task_ids=list(self.external_task_ids),
                    logical_dates=list(dttm_filter),
                    states=states,
                )
            if self.external_task_group_id:
                run_id_task_state_map = ti.get_task_states(
                    dag_id=self.external_dag_id,
                    task_group_id=self.external_task_group_id,
                    logical_dates=list(dttm_filter),
                )
                return _get_count_by_matched_states(run_id_task_state_map, states)
            return ti.get_dr_count(
                dag_id=self.external_dag_id,
                logical_dates=list(dttm_filter),
                states=states,
            )

        if self.failed_states:
            count = _get_count(self.failed_states)
            count_failed = self._calculate_count(count, dttm_filter)
            self._handle_failed_states(count_failed)

        if self.skipped_states:
            count = _get_count(self.skipped_states)
            count_skipped = self._calculate_count(count, dttm_filter)
            self._handle_skipped_states(count_skipped)

        count = _get_count(self.allowed_states)
        count_allowed = self._calculate_count(count, dttm_filter)
        return count_allowed == len(dttm_filter)

    def _calculate_count(self, count: int, dttm_filter: Sequence[datetime.datetime]) -> float | int:
        """Calculate the normalized count based on the type of check."""
        if self.external_task_ids:
            return count / len(self.external_task_ids)
        return count

    def _handle_failed_states(self, count_failed: float | int) -> None:
        """Handle failed states and raise appropriate exceptions."""
        if count_failed > 0:
            if self.external_task_ids:
                if self.soft_fail:
                    raise AirflowSkipException(
                        f"Some of the external tasks {self.external_task_ids} "
                        f"in DAG {self.external_dag_id} failed. Skipping due to soft_fail."
                    )
                raise ExternalTaskFailedError(
                    f"Some of the external tasks {self.external_task_ids} "
                    f"in DAG {self.external_dag_id} failed."
                )
            if self.external_task_group_id:
                if self.soft_fail:
                    raise AirflowSkipException(
                        f"The external task_group '{self.external_task_group_id}' "
                        f"in DAG '{self.external_dag_id}' failed. Skipping due to soft_fail."
                    )
                raise ExternalTaskGroupFailedError(
                    f"The external task_group '{self.external_task_group_id}' "
                    f"in DAG '{self.external_dag_id}' failed."
                )
            if self.soft_fail:
                raise AirflowSkipException(
                    f"The external DAG {self.external_dag_id} failed. Skipping due to soft_fail."
                )
            raise ExternalDagFailedError(f"The external DAG {self.external_dag_id} failed.")

    def _handle_skipped_states(self, count_skipped: float | int) -> None:
        """Handle skipped states and raise appropriate exceptions."""
        if count_skipped > 0:
            if self.external_task_ids:
                raise AirflowSkipException(
                    f"Some of the external tasks {self.external_task_ids} "
                    f"in DAG {self.external_dag_id} reached a state in our states-to-skip-on list. Skipping."
                )
            if self.external_task_group_id:
                raise AirflowSkipException(
                    f"The external task_group '{self.external_task_group_id}' "
                    f"in DAG {self.external_dag_id} reached a state in our states-to-skip-on list. Skipping."
                )
            raise AirflowSkipException(
                f"The external DAG {self.external_dag_id} reached a state in our states-to-skip-on list. "
                "Skipping."
            )

    if not AIRFLOW_V_3_0_PLUS:

        @provide_session
        def _poke_af2(self, dttm_filter: Sequence[datetime.datetime], session: Session = NEW_SESSION) -> bool:
            if self.check_existence and not self._has_checked_existence:
                self._check_for_existence(session=session)

            if self.failed_states:
                count_failed = self.get_count(dttm_filter, session, self.failed_states)
                self._handle_failed_states(count_failed)

            if self.skipped_states:
                count_skipped = self.get_count(dttm_filter, session, self.skipped_states)
                self._handle_skipped_states(count_skipped)

            count_allowed = self.get_count(dttm_filter, session, self.allowed_states)
            return count_allowed == len(dttm_filter)

    def execute(self, context: Context) -> None:
        """Run on the worker and defer using the triggers if deferrable is set to True."""
        if not self.deferrable:
            super().execute(context)
        else:
            dttm_filter = self._get_dttm_filter(context)
            if AIRFLOW_V_3_0_PLUS:
                self.defer(
                    timeout=self.execution_timeout,
                    trigger=WorkflowTrigger(
                        external_dag_id=self.external_dag_id,
                        external_task_group_id=self.external_task_group_id,
                        external_task_ids=self.external_task_ids,
                        allowed_states=self.allowed_states,
                        failed_states=self.failed_states,
                        skipped_states=self.skipped_states,
                        poke_interval=self.poll_interval,
                        soft_fail=self.soft_fail,
                        logical_dates=list(dttm_filter),
                        run_ids=None,
                        execution_dates=None,
                    ),
                    method_name="execute_complete",
                )
            else:
                self.defer(
                    timeout=self.execution_timeout,
                    trigger=WorkflowTrigger(
                        external_dag_id=self.external_dag_id,
                        external_task_group_id=self.external_task_group_id,
                        external_task_ids=self.external_task_ids,
                        allowed_states=self.allowed_states,
                        failed_states=self.failed_states,
                        skipped_states=self.skipped_states,
                        poke_interval=self.poll_interval,
                        soft_fail=self.soft_fail,
                        execution_dates=list(dttm_filter),
                        logical_dates=None,
                        run_ids=None,
                    ),
                    method_name="execute_complete",
                )

    def execute_complete(self, context: Context, event: dict[str, typing.Any] | None = None) -> None:
        """Execute when the trigger fires - return immediately."""
        if event is None:
            raise ExternalTaskNotFoundError("No event received from trigger")

        # Re-set as attribute after coming back from deferral - to be used by listeners
        self.external_dates_filter = self._serialize_dttm_filter(self._get_dttm_filter(context))

        if event["status"] == "success":
            self.log.info("External tasks %s has executed successfully.", self.external_task_ids)
        elif event["status"] == "skipped":
            raise AirflowSkipException("External job has skipped skipping.")
        elif event["status"] == "failed":
            if self.soft_fail:
                raise AirflowSkipException("External job has failed skipping.")
            raise ExternalDagFailedError("External job has failed.")
        else:
            if self.soft_fail:
                raise AirflowSkipException("External job has failed skipping.")
            raise ExternalTaskNotFoundError(
                "Error occurred while trying to retrieve task status. Please, check the "
                "name of executed task and Dag."
            )

    def _check_for_existence(self, session: Session) -> None:
        dag_to_wait = DagModel.get_current(self.external_dag_id, session)

        if not dag_to_wait:
            raise ExternalDagNotFoundError(f"The external DAG {self.external_dag_id} does not exist.")

        path = correct_maybe_zipped(dag_to_wait.fileloc)
        if not path or not os.path.exists(path):
            raise ExternalDagDeletedError(f"The external DAG {self.external_dag_id} was deleted.")

        if self.external_task_ids:
            refreshed_dag_info = DagBag(dag_to_wait.fileloc).get_dag(self.external_dag_id)
            if not refreshed_dag_info:
                raise ExternalDagNotFoundError(
                    f"The external DAG {self.external_dag_id} could not be loaded."
                )
            for external_task_id in self.external_task_ids:
                if not refreshed_dag_info.has_task(external_task_id):
                    raise ExternalTaskNotFoundError(
                        f"The external task {external_task_id} in DAG {self.external_dag_id} does not exist."
                    )

        if self.external_task_group_id:
            refreshed_dag_info = DagBag(dag_to_wait.fileloc).get_dag(self.external_dag_id)
            if not refreshed_dag_info:
                raise ExternalDagNotFoundError(
                    f"The external DAG {self.external_dag_id} could not be loaded."
                )
            if not refreshed_dag_info.has_task_group(self.external_task_group_id):
                raise ExternalTaskGroupNotFoundError(
                    f"The external task group '{self.external_task_group_id}' in "
                    f"DAG '{self.external_dag_id}' does not exist."
                )

        self._has_checked_existence = True

    def get_count(self, dttm_filter: Sequence[datetime.datetime], session: Session, states: list[str]) -> int:
        """
        Get the count of records against dttm filter and states.

        :param dttm_filter: date time filter for logical date
        :param session: airflow session object
        :param states: task or dag states
        :return: count of record against the filters
        """
        warnings.warn(
            "This method is deprecated and will be removed in future.", DeprecationWarning, stacklevel=2
        )
        return _get_count(
            dttm_filter,
            self.external_task_ids,
            self.external_task_group_id,
            self.external_dag_id,
            states,
            session,
        )

    def get_external_task_group_task_ids(
        self, session: Session, dttm_filter: Sequence[datetime.datetime]
    ) -> list[tuple[str, int]]:
        warnings.warn(
            "This method is deprecated and will be removed in future.", DeprecationWarning, stacklevel=2
        )
        if self.external_task_group_id is None:
            return []
        return _get_external_task_group_task_ids(
            list(dttm_filter), self.external_task_group_id, self.external_dag_id, session
        )

    def _get_logical_date(self, context: Context) -> datetime.datetime:
        """
        Handle backwards- and forwards-compatible retrieval of the date.

        to pass as the positional argument to execution_date_fn.
        """
        # Airflow 3.x: contexts define "logical_date" (or fall back to dag_run.run_after).
        if AIRFLOW_V_3_0_PLUS:
            logical_date = context.get("logical_date")
            dag_run = context.get("dag_run")
            if logical_date:
                return logical_date
            if dag_run and hasattr(dag_run, "run_after") and dag_run.run_after:
                return dag_run.run_after
            raise ValueError("Either `logical_date` or `dag_run.run_after` must be provided in the context")

        # Airflow 2.x and earlier: contexts used "execution_date"
        execution_date = context.get("execution_date")
        if not execution_date:
            raise ValueError("Either `execution_date` must be provided in the context`")
        if not isinstance(execution_date, datetime.datetime):
            raise ValueError("execution_date must be a datetime object")
        return execution_date

    def _handle_execution_date_fn(self, context: Context) -> datetime.datetime | list[datetime.datetime]:
        """
        Handle backward compatibility.

        This function is to handle backwards compatibility with how this operator was
        previously where it only passes the logical date, but also allow for the newer
        implementation to pass all context variables as keyword arguments, to allow
        for more sophisticated returns of dates to return.
        """
        from airflow.utils.operator_helpers import make_kwargs_callable

        # Remove "logical_date" because it is already a mandatory positional argument
        logical_date = self._get_logical_date(context)
        kwargs = {k: v for k, v in context.items() if k not in {"execution_date", "logical_date"}}
        # Add "context" in the kwargs for backward compatibility (because context used to be
        # an acceptable argument of execution_date_fn)
        kwargs["context"] = context
        if TYPE_CHECKING:
            assert self.execution_date_fn is not None
        kwargs_callable = make_kwargs_callable(self.execution_date_fn)
        return kwargs_callable(logical_date, **kwargs)


class ExternalTaskMarker(EmptyOperator):
    """
    Use this operator to indicate that a task on a different DAG depends on this task.

    When this task is cleared with "Recursive" selected, Airflow will clear the task on
    the other DAG and its downstream tasks recursively. Transitive dependencies are followed
    until the recursion_depth is reached.

    :param external_dag_id: The dag_id that contains the dependent task that needs to be cleared.
    :param external_task_id: The task_id of the dependent task that needs to be cleared.
    :param logical_date: The logical date of the dependent task execution that needs to be cleared.
    :param recursion_depth: The maximum level of transitive dependencies allowed. Default is 10.
        This is mostly used for preventing cyclic dependencies. It is fine to increase
        this number if necessary. However, too many levels of transitive dependencies will make
        it slower to clear tasks in the web UI.
    """

    template_fields = ["external_dag_id", "external_task_id", "logical_date"]
    if not AIRFLOW_V_3_0_PLUS:
        template_fields.append("execution_date")

    ui_color = "#4db7db"
    operator_extra_links = [ExternalDagLink()]

    # The _serialized_fields are lazily loaded when get_serialized_fields() method is called
    __serialized_fields: ClassVar[frozenset[str] | None] = None

    def __init__(
        self,
        *,
        external_dag_id: str,
        external_task_id: str,
        logical_date: str | datetime.datetime | None = "{{ logical_date.isoformat() }}",
        recursion_depth: int = 10,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.external_dag_id = external_dag_id
        self.external_task_id = external_task_id
        if isinstance(logical_date, datetime.datetime):
            self.logical_date = logical_date.isoformat()
        elif isinstance(logical_date, str):
            self.logical_date = logical_date
        else:
            raise TypeError(
                f"Expected str or datetime.datetime type for logical_date. Got {type(logical_date)}"
            )

        if not AIRFLOW_V_3_0_PLUS:
            self.execution_date = self.logical_date

        if recursion_depth <= 0:
            raise ValueError("recursion_depth should be a positive integer")
        self.recursion_depth = recursion_depth

    @classmethod
    def get_serialized_fields(cls):
        """Serialize ExternalTaskMarker to contain exactly these fields + templated_fields ."""
        if not cls.__serialized_fields:
            cls.__serialized_fields = frozenset(super().get_serialized_fields() | {"recursion_depth"})
        return cls.__serialized_fields



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from __future__ import annotations

import warnings
from collections.abc import Callable, Collection, Mapping, Sequence
from typing import TYPE_CHECKING, Any, ClassVar

from airflow.providers.common.compat.sdk import (
    DecoratedOperator,
    TaskDecorator,
    context_merge,
    task_decorator_factory,
)
from airflow.providers.standard.operators.bash import BashOperator
from airflow.sdk.definitions._internal.types import SET_DURING_EXECUTION
from airflow.utils.operator_helpers import determine_kwargs

if TYPE_CHECKING:
    from airflow.providers.common.compat.sdk import Context


class _BashDecoratedOperator(DecoratedOperator, BashOperator):
    """
    Wraps a Python callable and uses the callable return value as the Bash command to be executed.

    :param python_callable: A reference to an object that is callable.
    :param op_kwargs: A dictionary of keyword arguments that will get unpacked
        in your function (templated).
    :param op_args: A list of positional arguments that will get unpacked when
        calling your callable (templated).
    """

    template_fields: Sequence[str] = (*DecoratedOperator.template_fields, *BashOperator.template_fields)
    template_fields_renderers: ClassVar[dict[str, str]] = {
        **DecoratedOperator.template_fields_renderers,
        **BashOperator.template_fields_renderers,
    }

    custom_operator_name: str = "@task.bash"
    overwrite_rtif_after_execution: bool = True

    def __init__(
        self,
        *,
        python_callable: Callable,
        op_args: Collection[Any] | None = None,
        op_kwargs: Mapping[str, Any] | None = None,
        **kwargs,
    ) -> None:
        if kwargs.pop("multiple_outputs", None):
            warnings.warn(
                f"`multiple_outputs=True` is not supported in {self.custom_operator_name} tasks. Ignoring.",
                UserWarning,
                stacklevel=3,
            )

        super().__init__(
            python_callable=python_callable,
            op_args=op_args,
            op_kwargs=op_kwargs,
            bash_command=SET_DURING_EXECUTION,
            multiple_outputs=False,
            **kwargs,
        )

    def execute(self, context: Context) -> Any:
        context_merge(context, self.op_kwargs)
        kwargs = determine_kwargs(self.python_callable, self.op_args, context)

        self.bash_command = self.python_callable(*self.op_args, **kwargs)

        if not isinstance(self.bash_command, str) or self.bash_command.strip() == "":
            raise TypeError("The returned value from the TaskFlow callable must be a non-empty string.")

        self._is_inline_cmd = self._is_inline_command(bash_command=self.bash_command)
        self.render_template_fields(context)
        return super().execute(context)


def bash_task(
    python_callable: Callable | None = None,
    **kwargs,
) -> TaskDecorator:
    """
    Wrap a function into a BashOperator.

    Accepts kwargs for operator kwargs. Can be reused in a single DAG. This function is only used only used
    during type checking or auto-completion.

    :param python_callable: Function to decorate.

    :meta private:
    """
    return task_decorator_factory(
        python_callable=python_callable,
        decorated_operator_class=_BashDecoratedOperator,
        **kwargs,
    )



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
"""Utilities for creating a virtual environment."""

from __future__ import annotations

import logging
import os
import shlex
import shutil
import subprocess
import warnings
from pathlib import Path

import jinja2
from jinja2 import select_autoescape

from airflow.providers.common.compat.sdk import conf


def _is_uv_installed() -> bool:
    """
    Verify whether the uv tool is installed by checking if it's included in the system PATH or installed as a package.

    :return: True if it is. Whichever way of checking it works, is fine.
    """
    return bool(shutil.which("uv"))


def _use_uv() -> bool:
    """
    Check if the uv tool should be used.

    :return: True if uv should be used.
    """
    venv_install_method = conf.get("standard", "venv_install_method", fallback="auto").lower()
    if venv_install_method == "auto":
        return _is_uv_installed()
    if venv_install_method == "uv":
        return True
    return False


def _generate_uv_cmd(tmp_dir: str, python_bin: str, system_site_packages: bool) -> list[str]:
    """Build the command to install the venv via UV."""
    if python_bin == "python" or python_bin == "python3":
        python_interpreter_exists = bool(shutil.which(python_bin))
        if not python_interpreter_exists:
            warnings.warn(
                f"uv trying to use `{python_bin}` as the python interpreter. it could lead to errors if the python interpreter not found in PATH. "
                f"please specify python_version in operator.",
                UserWarning,
                stacklevel=3,
            )
    cmd = ["uv", "venv", "--allow-existing", "--seed", "--python", python_bin]
    if system_site_packages:
        cmd.append("--system-site-packages")
    cmd.append(tmp_dir)
    return cmd


def _generate_venv_cmd(tmp_dir: str, python_bin: str, system_site_packages: bool) -> list[str]:
    """We are using venv command instead of venv module to allow creation of venv for different python versions."""
    cmd = [python_bin, "-m", "venv", tmp_dir]
    if system_site_packages:
        cmd.append("--system-site-packages")
    return cmd


def _generate_uv_install_cmd_from_file(
    tmp_dir: str, requirements_file_path: str, pip_install_options: list[str]
) -> list[str]:
    return [
        "uv",
        "pip",
        "install",
        "--python",
        f"{tmp_dir}/bin/python",
        *pip_install_options,
        "-r",
        requirements_file_path,
    ]


def _generate_pip_install_cmd_from_file(
    tmp_dir: str, requirements_file_path: str, pip_install_options: list[str]
) -> list[str]:
    return [f"{tmp_dir}/bin/pip", "install", *pip_install_options, "-r", requirements_file_path]


def _generate_uv_install_cmd_from_list(
    tmp_dir: str, requirements: list[str], pip_install_options: list[str]
) -> list[str]:
    return ["uv", "pip", "install", "--python", f"{tmp_dir}/bin/python", *pip_install_options, *requirements]


def _generate_pip_install_cmd_from_list(
    tmp_dir: str, requirements: list[str], pip_install_options: list[str]
) -> list[str]:
    return [f"{tmp_dir}/bin/pip", "install", *pip_install_options, *requirements]


def _generate_pip_conf(conf_file: Path, index_urls: list[str]) -> None:
    if index_urls:
        pip_conf_options = f"index-url = {index_urls[0]}"
        if len(index_urls) > 1:
            pip_conf_options += f"\nextra-index-url = {' '.join(x for x in index_urls[1:])}"
    else:
        pip_conf_options = "no-index = true"
    conf_file.write_text(f"[global]\n{pip_conf_options}")


def _index_urls_to_uv_env_vars(index_urls: list[str] | None = None) -> dict[str, str]:
    uv_index_env_vars = {}
    if index_urls:
        uv_index_env_vars = {"UV_DEFAULT_INDEX": index_urls[0]}
        if len(index_urls) > 1:
            uv_index_env_vars["UV_INDEX"] = " ".join(x for x in index_urls[1:])
    return uv_index_env_vars


def _execute_in_subprocess(cmd: list[str], cwd: str | None = None, env: dict[str, str] | None = None) -> None:
    """
    Execute a process and stream output to logger.

    :param cmd: command and arguments to run
    :param cwd: Current working directory passed to the Popen constructor
    :param env: Additional environment variables to set for the subprocess.
    """
    log = logging.getLogger(__name__)

    log.info("Executing cmd: %s", " ".join(shlex.quote(c) for c in cmd))
    with subprocess.Popen(
        cmd,
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
        bufsize=0,
        close_fds=False,
        cwd=cwd,
        env=env,
    ) as proc:
        log.info("Output:")
        if proc.stdout:
            with proc.stdout:
                for line in iter(proc.stdout.readline, b""):
                    log.info("%s", line.decode().rstrip())

        exit_code = proc.wait()
    if exit_code != 0:
        raise subprocess.CalledProcessError(exit_code, cmd)


def prepare_virtualenv(
    venv_directory: str,
    python_bin: str,
    system_site_packages: bool,
    requirements: list[str] | None = None,
    requirements_file_path: str | None = None,
    pip_install_options: list[str] | None = None,
    index_urls: list[str] | None = None,
) -> str:
    """
    Create a virtual environment and install the additional python packages.

    :param venv_directory: The path for directory where the environment will be created.
    :param python_bin: Path to the Python executable.
    :param system_site_packages: Whether to include system_site_packages in your virtualenv.
        See virtualenv documentation for more information.
    :param requirements: List of additional python packages.
    :param requirements_file_path: Path to the ``requirements.txt`` file.
    :param pip_install_options: a list of pip install options when installing requirements
        See 'pip install -h' for available options
    :param index_urls: an optional list of index urls to load Python packages from.
        If not provided the system pip conf will be used to source packages from.
    :return: Path to a binary file with Python in a virtual environment.
    """
    if pip_install_options is None:
        pip_install_options = []

    if requirements is not None and requirements_file_path is not None:
        raise ValueError("Either requirements OR requirements_file_path has to be passed, but not both")

    if index_urls is not None:
        _generate_pip_conf(Path(venv_directory) / "pip.conf", index_urls)

    if _use_uv():
        venv_cmd = _generate_uv_cmd(venv_directory, python_bin, system_site_packages)
        _execute_in_subprocess(venv_cmd, env={**os.environ, **_index_urls_to_uv_env_vars(index_urls)})
    else:
        venv_cmd = _generate_venv_cmd(venv_directory, python_bin, system_site_packages)
        _execute_in_subprocess(venv_cmd)

    pip_cmd = None
    if requirements is not None and len(requirements) != 0:
        if _use_uv():
            pip_cmd = _generate_uv_install_cmd_from_list(venv_directory, requirements, pip_install_options)
        else:
            pip_cmd = _generate_pip_install_cmd_from_list(venv_directory, requirements, pip_install_options)
    if requirements_file_path is not None and requirements_file_path:
        if _use_uv():
            pip_cmd = _generate_uv_install_cmd_from_file(
                venv_directory, requirements_file_path, pip_install_options
            )
        else:
            pip_cmd = _generate_pip_install_cmd_from_file(
                venv_directory, requirements_file_path, pip_install_options
            )

    if pip_cmd:
        _execute_in_subprocess(pip_cmd, env={**os.environ, **_index_urls_to_uv_env_vars(index_urls)})

    return f"{venv_directory}/bin/python"


def write_python_script(
    jinja_context: dict,
    filename: str,
    render_template_as_native_obj: bool = False,
):
    """
    Render the python script to a file to execute in the virtual environment.

    :param jinja_context: The jinja context variables to unpack and replace with its placeholders in the
        template file.
    :param filename: The name of the file to dump the rendered script to.
    :param render_template_as_native_obj: If ``True``, rendered Jinja template would be converted
        to a native Python object
    """
    template_loader = jinja2.FileSystemLoader(searchpath=os.path.dirname(__file__))
    template_env: jinja2.Environment
    if render_template_as_native_obj:
        template_env = jinja2.nativetypes.NativeEnvironment(
            loader=template_loader, undefined=jinja2.StrictUndefined
        )
    else:
        template_env = jinja2.Environment(
            loader=template_loader,
            undefined=jinja2.StrictUndefined,
            autoescape=select_autoescape(["html", "xml"]),
        )
    template = template_env.get_template("python_virtualenv_script.jinja2")
    template.stream(**jinja_context).dump(filename)
